import re

from .baseobject import BaseObject


class NodeParams(BaseObject):
    keys = {
        "index": "index",
        "network": "network",
        "api_password": "api_password",
        "network_port": "network_port",
        "session_port": "session_port",
        "module_address": "module_address",
        "safe_address": "safe_address",
        "identity": "identity",
        "identity_password": "identity_password",
        "folder": "folder",
        "surname": "surname",
        "blokli_url": "blokli_url",
    }

    @property
    def network_name(self):
        return self.network.meta.name

    @property
    def network_port(self):
        return self.network.meta.network_port_base + self.index

    @property
    def session_port(self):
        return self.network.meta.session_port_base + self.index

    @property
    def filename(self):
        return f"hoprd-{self.network_name}-{self.node_suffix}"

    @property
    def node_suffix(self):
        surname = getattr(self, "surname", None)
        if surname:
            slug = re.sub(r"[^0-9A-Za-z]+", "-", surname).strip("-").lower()
            if slug:
                return slug
        return str(self.index)

    @property
    def as_dict(self):
        data = super().as_dict
        data["node_suffix"] = self.node_suffix
        if getattr(self, "surname", None):
            data["surname"] = self.surname
        return data

    @property
    def config_folder(self):
        folder = self.folder.joinpath(".hopr-configs", self.network_name)
        folder.mkdir(parents=True, exist_ok=True)
        return folder

    @property
    def id_folder(self):
        folder = self.folder.joinpath(".hopr-ids", self.network_name)
        folder.mkdir(parents=True, exist_ok=True)
        return folder

    @property
    def config_file(self):
        return self.config_folder.joinpath(f"{self.filename}.cfg.yaml")

    @property
    def id_file(self):
        return self.id_folder.joinpath(f"{self.filename}.id")



import logging
import subprocess
from pathlib import Path

import click

try:
    import tomllib as toml_loader
except ModuleNotFoundError:  # pragma: no cover - Python < 3.11
    import tomli as toml_loader

import yaml

from .config_filling import ConfigFilling
from .docker_compose import DockerComposeGenerator
from .library import get_template, replace_fields
from .network import Network
from .params import NodeParams
from .yaml import Aggregating, AutoFunding, AutoRedeeming, ClosureFinalizer, IPv4, Token

YAML_OBJECTS = [IPv4, Token, Aggregating,
                AutoFunding, AutoRedeeming, ClosureFinalizer]
YAML_TAG_MAP = {cls.__name__: cls for cls in YAML_OBJECTS}

logging.basicConfig(level=logging.DEBUG,
                    format="%(levelname)-8s:%(message)s", datefmt="[%X]")
logger = logging.getLogger("hoprd-config-generator")


def _expand_tagged_sections(value):
    """Convert TOML dotted tables into tag-aware dictionaries.

    When the params TOML file expresses YAML tags via dotted array tables
    (for example ``[[config.hopr.strategy.strategies.ClosureFinalizer]]``), the
    parsed structure ends up as a mapping whose keys correspond to the class
    names (``{"ClosureFinalizer": [{...}]}``).  The configuration generator
    expects the previous ``{"tag": "ClosureFinalizer", ...}`` layout, so we
    normalize those mappings back into a list of dictionaries that carry the
    ``tag`` attribute.  Regular dictionaries are processed recursively without
    modification.
    """

    if isinstance(value, dict):
        normalized = {k: _expand_tagged_sections(v) for k, v in value.items()}

        if normalized and all(key in YAML_TAG_MAP for key in normalized):
            expanded = []
            for tag_name, entries in normalized.items():
                entries_list = entries if isinstance(entries, list) else [entries]
                for entry in entries_list:
                    if isinstance(entry, dict):
                        payload = dict(entry)
                    else:
                        payload = {"value": entry}
                    payload.setdefault("tag", tag_name)
                    expanded.append(payload)
            return expanded

        return normalized

    if isinstance(value, list):
        return [_expand_tagged_sections(v) for v in value]

    return value


def _normalize_nodes_section(config_content):
    """Allow ``[[nodes.<surname>]]`` tables to define per-node entries."""

    nodes = config_content.get("nodes")
    if isinstance(nodes, dict):
        normalized_nodes = []
        for surname, entries in nodes.items():
            if isinstance(entries, list):
                entries_list = entries
            else:
                entries_list = [entries]

            for entry in entries_list:
                if not isinstance(entry, dict):
                    raise TypeError(
                        "Node definitions must be tables; "
                        f"got {type(entry).__name__} for '{surname}'"
                    )

                payload = dict(entry)
                if surname not in (None, ""):
                    payload.setdefault("surname", surname)
                normalized_nodes.append(payload)

        config_content["nodes"] = normalized_nodes

    return config_content


def _instantiate_tagged_objects(value):
    if isinstance(value, dict):
        tag_name = value.get("tag")
        if isinstance(tag_name, str):
            tag_name = tag_name.lstrip("!")
            yaml_class = YAML_TAG_MAP.get(tag_name)
            if yaml_class is None:
                raise ValueError(f"Unknown tag '{tag_name}' in params file")

            payload = {
                key: _instantiate_tagged_objects(val)
                for key, val in value.items()
                if key != "tag"
            }

            annotations = list(getattr(yaml_class, "__annotations__", {}).keys())
            if "value" in payload and annotations:
                first_attr = annotations[0]
                value_payload = payload.pop("value")
                if first_attr not in payload:
                    payload[first_attr] = value_payload

            return yaml_class(**payload)

        return {k: _instantiate_tagged_objects(v) for k, v in value.items()}

    if isinstance(value, list):
        return [_instantiate_tagged_objects(v) for v in value]

    return value


@click.command()
@click.option("--params", "params_file", type=click.Path(path_type=Path))
@click.option("--folder", "base_folder", default=Path("./.hoprd-nodes"), type=click.Path(path_type=Path))
def main(params_file: Path, base_folder: Path):
    for cls in YAML_OBJECTS:
        yaml.SafeLoader.add_constructor(cls.yaml_tag, cls.from_yaml)
        yaml.SafeDumper.add_multi_representer(cls, cls.to_yaml)

    node_template = get_template(Path("node.yaml"))
    logger.info(f"Successfuly read node config file template")

    try:
        ip_addr = subprocess.check_output(
            "curl -s https://ipinfo.io/ip", shell=True).decode().strip()
        logger.info(f"Retrieved machine IP as '{ip_addr}'")
    except subprocess.CalledProcessError as exc:
        ip_addr = "0.0.0.0"
        logger.warning(
            "Could not retrieve public IP (defaulting to %s): %s", ip_addr, exc)

    with params_file.open("rb") as f:
        config_content: dict = toml_loader.load(f)

    config_content = _expand_tagged_sections(config_content)
    config_content = _instantiate_tagged_objects(config_content)
    config_content = _normalize_nodes_section(config_content)

    network = Network(config_content)
    logger.info(f"Loaded {len(network.nodes)} nodes for '{network.meta.name}' network")

    # Create list of nodes
    nodes_params = list[NodeParams]()
    for index, node in enumerate(network.nodes, 1):
        params = {
            "index": index,
            "network": network,
            "folder": base_folder,
            "blokli_url": network.meta.blokli_url,
        }
        node_param = NodeParams(params | node.as_dict)
        nodes_params.append(node_param)

    # Generate config files
    logger.info("Generating config files")
    for obj in nodes_params:
        config = replace_fields(node_template, obj.network.config)
        config = ConfigFilling.apply(config, obj, ip_addr=ip_addr)

        with open(obj.config_file, "w") as f:
            yaml.safe_dump(config, f)
            logger.debug(
                f"  Config for {obj.filename} at '{obj.config_file}'")

    # Generate identity files
    logger.info("Generating identity files")
    for obj in nodes_params:
        with open(obj.id_file, "w") as f:
            f.write(obj.identity)
            logger.debug(
                f"  Identity for {obj.filename} at '{obj.id_file}'")

    # Generate docker compose file
    logger.info("Generating docker-compose file")
    compose_output = Path(f"./docker-compose.{network.meta.name}.yml")
    compose_generator = DockerComposeGenerator()
    compose_generator.write(
        compose_output, network=network, nodes=nodes_params, config_content=config_content
    )
    logger.info(f"Docker compose file at '{compose_output}'")


if __name__ == "__main__":
    main()



from .baseobject import BaseObject


class Node(BaseObject):
    keys = {
        "safe": "safe_address",
        "module": "module_address",
        "api_password": "api_password",
        "identity_password": "identity_password",
        "identity": "identity",
        "address": "node_address",
        "surname": "surname",
    }



import yaml


class YAMLParser(yaml.YAMLObject):
    def __init__(self, **kwargs):
        params = list(vars(self.__class__)['__annotations__'].keys())
        kwargs = {k: v for k, v in kwargs.items() if k in params}

        for key, value in kwargs.items():
            setattr(self, key, value)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.yaml_tag = u'!{}'.format(cls.__name__)

    def __repr__(self):
        vars = self.__dict__
        var_string = ', '.join(
            [f"{key}={value}" for key, value in vars.items()])
        return f"{self.__class__.__name__}({var_string})"

    @classmethod
    def from_yaml(cls, loader, node):
        if isinstance(node.value, list):
            return cls(**loader.construct_mapping(node))
        else:
            unique_param = list(vars(cls)['__annotations__'].keys())[0]
            return cls(**{unique_param: node.__dict__["value"]})

    @classmethod
    def to_yaml(cls, dumper, data):
        if hasattr(cls, "scalar") and cls.scalar is True:
            return dumper.represent_scalar(cls.yaml_tag, list(data.__dict__.values())[0])
        else:
            return dumper.represent_mapping(cls.yaml_tag, data.__dict__)



from .parser import YAMLParser


class IPv4(YAMLParser):
    scalar = True
    ipv4: str

class Token(YAMLParser):
    scalar = True
    token: str

class Aggregating(YAMLParser):
    aggregation_threshold: int
    unrealized_balance_ratio: float
    aggregate_on_channel_close: bool

class AutoFunding(YAMLParser):
    funding_amount: str
    min_stake_threshold: str

class AutoRedeeming(YAMLParser):
    redeem_only_aggregated: bool
    redeem_all_on_close: bool
    minimum_redeem_ticket_value: str
    redeem_on_winning: bool

class ClosureFinalizer(YAMLParser):
    max_closure_overdue: int


from .baseobject import BaseObject


class Meta(BaseObject):
    keys = {
        "version": "version",
        "name": "name",
        "api_port_base": "api_port_base",
        "network_port_base": "network_port_base",
        "session_port_base": "session_port_base",
        "blokli_url": "blokli_url",
    }


import collections.abc
from copy import deepcopy
from pathlib import Path
from typing import Any

import yaml
from jinja2 import Template


def convert(value: Any):
    if value is None:
        return None

    try:
        value = float(value)
    except ValueError:
        pass
    except TypeError:
        pass

    try:
        integer = int(value)
        if integer == value:
            value = integer
    except ValueError:
        pass
    except TypeError:
        pass

    return value


def get_template(filename: Path, template_folder: str = "templates"):
    template_path = Path(__file__).absolute(
    ).parent.joinpath(template_folder, filename)

    if filename.suffix in [".yaml", ".yml"]:
        with open(template_path, "r") as f:
            return yaml.safe_load(f)

    elif filename.suffix in [".j2"]:
        with open(template_path, "r") as f:
            return Template(f.read(), trim_blocks=True, lstrip_blocks=True)

    else:
        return filename


def replace_fields(source: dict, addons: dict) -> dict:
    copy = deepcopy(source)

    for k, v in addons.items():
        if isinstance(v, collections.abc.Mapping):
            copy[k] = replace_fields(copy.get(k, {}), v)
        else:
            copy[k] = v
    return copy


def set_nested_value(d, keys, value):
    for key in keys[:-1]:
        d = d.setdefault(key, {})
    d[keys[-1]] = value


def set_nested_attr(d, keys, attr, value):
    for key in keys:
        d = d.setdefault(key, {})
    setattr(d, attr, value)


def get_nested_value(d, keys):
    for key in keys:
        d = getattr(d, key)
    return d



from .baseobject import BaseObject
from .meta import Meta
from .node import Node


class Network(BaseObject):
    keys = {
        "meta": "meta",
        "config": "config",
        "nodes": "nodes",
        "slug": "nodes_slug",
        "networks": "networks",
    }

    def post_init(self):
        self.nodes = [Node(subdata) for subdata in self.nodes]
        self.meta = Meta(self.meta)



from typing import Any

from .library import get_nested_value, set_nested_attr, set_nested_value


class ConfigFilling:
    keys = {
        "api/auth.token": "api_password",
        "hopr/host/port": "network_port",
        "hopr/safe_module/module_address": "module_address",
        "hopr/safe_module/safe_address": "safe_address",
        "identity/password": "identity_password",
        "hopr/host/address.ipv4": "*ip_addr",
        "blokli_url": "blokli_url",
    }

    @classmethod
    def apply(cls, dictionary: dict, object: Any, **kwargs):
        for key, value in cls.keys.items():
            splits = key.split('.')

            if value.startswith("*"):
                parsed_value = kwargs[value.lstrip('*')]
            else:
                parsed_value = get_nested_value(object, value.split("/"))

            path = splits[0].split('/')

            if len(splits) == 1:
                set_nested_value(dictionary, path, parsed_value)
            else:
                path = splits[0].split('/')
                set_nested_attr(dictionary, path, splits[1], parsed_value)

        return dictionary



"""Utilities for rendering docker-compose files.

This module centralizes the logic required to prepare the context passed to the
``docker-compose.yml.j2`` template.  Previously this logic lived inside the
``__main__`` module which made it difficult to reason about and to test.  The
new structure provides small, well-named helpers that describe the intent of
each operation (parsing networks, formatting YAML, rendering/writing the
compose file, ...).  This makes the process easier to maintain and improves the
readability of ``__main__``.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from textwrap import indent
from typing import Iterable, Mapping, Sequence

import yaml

from .library import get_template
from .network import Network
from .params import NodeParams


@dataclass(frozen=True)
class DockerComposeContext:
    """All of the information required by the docker-compose template."""

    services: Sequence[dict]
    version: str | None
    network: str
    envvars: Mapping[str, object] = field(default_factory=dict)
    networks: Mapping[str, object] = field(default_factory=dict)

    @classmethod
    def from_config(
        cls,
        network: Network,
        nodes: Iterable[NodeParams],
        config_content: Mapping[str, object],
    ) -> "DockerComposeContext":
        """Create a context from the parsed network and raw config data."""

        networks = dict(getattr(network, "networks", None) or {})
        return cls(
            services=[node.as_dict for node in nodes],
            version=network.meta.version,
            network=network.meta.name,
            envvars=dict(config_content.get("env", {})),
            networks=networks,
        )

    @property
    def network_names(self) -> tuple[str, ...]:
        return tuple(self.networks.keys())

    @property
    def networks_yaml(self) -> str:
        if not self.networks:
            return ""
        dumped = yaml.safe_dump(self.networks, sort_keys=False)
        return indent(dumped, "  ")

    def to_template_kwargs(self) -> dict:
        """Return a dictionary ready to be passed to the Jinja template."""

        return {
            "services": self.services,
            "version": self.version,
            "network": self.network,
            "envvars": self.envvars,
            "networks": self.networks or None,
            "networks_yaml": self.networks_yaml,
            "network_names": self.network_names,
        }


class DockerComposeGenerator:
    """Helper responsible for rendering docker-compose files."""

    def __init__(self, template: Path | str = Path("docker-compose.yml.j2")):
        self.template_path = Path(template)

    def _render_template(self, context: DockerComposeContext) -> str:
        template = get_template(self.template_path)
        return template.render(**context.to_template_kwargs())

    def render(
        self,
        network: Network,
        nodes: Sequence[NodeParams],
        config_content: Mapping[str, object],
    ) -> str:
        context = DockerComposeContext.from_config(network, nodes, config_content)
        return self._render_template(context)

    def write(
        self,
        output: Path,
        network: Network,
        nodes: Sequence[NodeParams],
        config_content: Mapping[str, object],
    ) -> Path:
        """Render the compose file and write it to ``output``."""

        content = self.render(network, nodes, config_content)
        output_path = Path(output)
        output_path.write_text(content)
        return output_path



from .library import convert


class BaseObject:
    def __init__(self, data: dict, **kwargs):
        for key, value in self.keys.items():
            v = data
            for subkey in value.split("/"):
                v = v.get(subkey, None)
                if v is None:
                    continue
            try:
                setattr(self, key, convert(v))
            except:
                pass

        for key, value in kwargs.items():
            setattr(self, key, value)

        self.post_init()

    def post_init(self):
        pass

    @property
    def is_null(self):
        return all(getattr(self, key) is None for key in self.keys.keys())

    @property
    def as_dict(self) -> dict:
        return {value: getattr(self, key) for key, value in self.keys.items()}

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return all(
            getattr(self, key) == getattr(other, key) for key in self.keys.keys()
        )
